/*
 * DMA support for the JZ4780.
 *
 * Copyright (C) 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>
#include <stdint.h>



/* Enumerated types for various transfer parameters. */

enum Dma_jz4780_request_type
{
  Dma_request_i2s1_out  = 4,
  Dma_request_i2s1_in   = 5,
  Dma_request_i2s0_out  = 6,
  Dma_request_i2s0_in   = 7,
  Dma_request_auto      = 8,
  Dma_request_sadc_in   = 9,
  Dma_request_uart4_out = 12,
  Dma_request_uart4_in  = 13,
  Dma_request_uart3_out = 14,
  Dma_request_uart3_in  = 15,
  Dma_request_uart2_out = 16,
  Dma_request_uart2_in  = 17,
  Dma_request_uart1_out = 18,
  Dma_request_uart1_in  = 19,
  Dma_request_uart0_out = 20,
  Dma_request_uart0_in  = 21,
  Dma_request_ssi0_out  = 22,
  Dma_request_ssi0_in   = 23,
  Dma_request_ssi1_out  = 24,
  Dma_request_ssi1_in   = 25,
  Dma_request_msc0_out  = 26,
  Dma_request_msc0_in   = 27,
  Dma_request_msc1_out  = 28,
  Dma_request_msc1_in   = 29,
  Dma_request_msc2_out  = 30,
  Dma_request_msc2_in   = 31,
  Dma_request_pcm0_out  = 32,
  Dma_request_pcm0_in   = 33,
  Dma_request_i2c0_out  = 36,
  Dma_request_i2c0_in   = 37,
  Dma_request_i2c1_out  = 38,
  Dma_request_i2c1_in   = 39,
  Dma_request_i2c2_out  = 40,
  Dma_request_i2c2_in   = 41,
  Dma_request_i2c3_out  = 42,
  Dma_request_i2c3_in   = 43,
  Dma_request_i2c4_out  = 44,
  Dma_request_i2c4_in   = 45,
  Dma_request_des_out   = 46,
  Dma_request_des_in    = 47,
};

/* Descriptor structure. */

struct jz4780_dma_descriptor
{
  uint32_t command, source, destination, transfer_count,
           stride, request_source, reserved0, reserved1;
};



#ifdef __cplusplus

#include <l4/devices/dma-generic.h>
#include <l4/devices/hw_mmio_register_block.h>

// Forward declaration.

class Dma_jz4780_chip;



// DMA channel.

class Dma_jz4780_channel : public Dma_channel
{
private:
  Hw::Register_block<32> _regs;
  Dma_chip *_chip;
  uint8_t _channel;
  l4_cap_idx_t _irq = L4_INVALID_CAP;

public:
  Dma_jz4780_channel(Dma_chip *chip, uint8_t channel, l4_addr_t start,
                     l4_cap_idx_t irq = L4_INVALID_CAP);

  unsigned int transfer(uint32_t source, uint32_t destination,
                        unsigned int count,
                        bool source_increment, bool destination_increment,
                        uint8_t source_width, uint8_t destination_width,
                        uint8_t transfer_unit_size,
                        int type=Dma_request_auto,
                        l4_addr_t desc_vaddr = 0,
                        l4re_dma_space_dma_addr_t desc_paddr = 0);

  unsigned int wait();

protected:
  // Transfer property configuration.

  uint32_t encode_req_detect_int_length(uint8_t units);

  uint32_t encode_source_port_width(uint8_t width);

  uint32_t encode_destination_port_width(uint8_t width);

  uint32_t encode_transfer_unit_size(uint8_t size);

  // Transaction control.

  void ack_irq();

  bool completed();

  bool error();

  bool halted();

  bool wait_for_irq();

  bool wait_for_irq(unsigned int timeout);
};

// DMA device control.

class Dma_jz4780_chip : public Dma_chip
{
private:
  Hw::Register_block<32> _regs;
  l4_addr_t _start, _end;
  Cpm_chip *_cpm;

public:
  Dma_jz4780_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm);

  void disable();

  void enable();

  Dma_channel *get_channel(uint8_t channel,
                                  l4_cap_idx_t irq = L4_INVALID_CAP);

  // Transaction control.

  void ack_irq(uint8_t channel);

  bool error();

  bool halted();

  bool have_interrupt(uint8_t channel);

  // Descriptor operations.

  void commit_descriptor(uint8_t channel);
};

Dma_chip *jz4780_dma_chip(l4_addr_t start, l4_addr_t end, Cpm_chip *cpm);

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4780_dma_init(l4_addr_t start, l4_addr_t end, void *cpm);

void jz4780_dma_disable(void *dma_chip);

void jz4780_dma_enable(void *dma_chip);

void *jz4780_dma_get_channel(void *dma, uint8_t channel, l4_cap_idx_t irq);

unsigned int jz4780_dma_transfer(void *dma_channel,
                                 uint32_t source, uint32_t destination,
                                 unsigned int count,
                                 int source_increment, int destination_increment,
                                 uint8_t source_width, uint8_t destination_width,
                                 uint8_t transfer_unit_size,
                                 enum Dma_jz4780_request_type type);

unsigned int jz4780_dma_transfer_descriptor(void *dma_channel,
                                 uint32_t source, uint32_t destination,
                                 unsigned int count,
                                 int source_increment, int destination_increment,
                                 uint8_t source_width, uint8_t destination_width,
                                 uint8_t transfer_unit_size,
                                 enum Dma_jz4780_request_type type,
                                 l4_addr_t desc_vaddr,
                                 l4re_dma_space_dma_addr_t desc_paddr);

unsigned int jz4780_dma_wait(void *dma_channel);

EXTERN_C_END
